#!/usr/bin/perl -w
#
# This script creates a "master table of contents" for a set of LyX docs.
# It does so by going through the files and printing out all of the
# chapter, section, and sub(sub)section headings out. (It numbers the
# sections sequentially; hopefully noone's using Section* in the docs.)
#
# You should just be able to say "Doc_toc.pl > TOC.lyx", and it will
# run on all the files listed in @ARGV a couple lines below. Alternatively,
# you can explicitly give it a set of files to run on. (If your perl isn't
# in /usr/bin/perl, you can say "perl Doc_toc.pl [files.lyx...] > TOC.lyx")

# Add (whitespace separated) "Joe" to this list to create a TOC for Joe.lyx
@ARGV = map {"$_.lyx"} qw 
    (Intro Tutorial UserGuide Extended Customization FAQ Reference) 
    unless @ARGV;

use strict;

my $printing = 0; # are we currently supposed to be copying input to STDOUT?
my @LyX_classes = qw (Chapter Section Subsection Subsubsection);

# First print out initial info
print <<"ENDPREAMBLE";
#This file was created by The LyX Team Robot
#LyX 1.0 (C) 1995-1999 Matthias Ettrich and the LyX Team
\\lyxformat 2.15
\\textclass article

\\layout Title
LyX Doc Table of Contents

ENDPREAMBLE

my @matches;
my $level;
my $match = "";
my $sub_match;
my $ignore = 0; # ignore lines (e.g. footnote in a heading)
my @sec_counter;
# Now loop through files
while (<>) {
    # first few lines of the file
    unless (@matches) {
	if (/^\\textclass (\w+)/) {
	    # Hopefully it's book, report, or article
	    my $class = $1;
	    @matches = @LyX_classes;
	    # Article doesn't have Chapters
	    if ($class eq "article") {
	        shift @matches;
	    }
	    $match = $matches[0];
	    $sub_match = $matches[1];
	}

	next;
    }

    # Footnotes in a section heading could confuse things!
    # And don't bother printing out section labels.
    if (/^\\begin_float footnote/ || /^\\begin_inset LatexCommand \\label/) {
        $ignore = 1;
    } 
    # Unset ignoring. But note that end_inset could be the end of another
    # kind of inset!
    if ($ignore && /^\\end_(float|inset)/) {
        $ignore = 0;
	next; # don't print out \end_float line
    }
    next if $ignore;

    # Now actually handle title & section headings
    if (/^\\layout\s+(\w+)/) {
        my $layout = $1;
	my $found = 0; # did we find the start of a heading?
    
	if ($layout eq "Title") {
	    $found = 1;
	    @sec_counter = scalar(@matches) x (0); # (re)start section counters
	    print "\\layout Section*\n";

	} else {
	    foreach my $level (0 .. $#matches) {
		if ($layout eq $matches[$level]) {

		    # If this is the first subsection of this level,
		    # start nesting. Top level (level 0) needs no nesting
		    if ($level && !$sec_counter[$level]) {
			    print "\\begin_deeper\n";
		    }

		    # A new section ends any sub- or subsubsections
		    # below it, etc.
		    # (Make sure to do this before calling &section_number!
		    foreach my $sublevel ($level+1 .. $#sec_counter) {
			if ($sec_counter[$sublevel]) {
			    print "\\end_deeper\n";
			    $sec_counter[$sublevel] = 0;
			}
		    }

		    $found = 1;
		    $sec_counter[$level]++;
		    my $sec = &section_number (@sec_counter);
		    print "\\layout Description\n$sec ";
		    last; # don't need to try any more matches
		}
	    }
	}

	# If we found "Title" or one of the section layouts, then we're
	# in a section header. Otherwise, this \layout command ENDS any section
	# we might have been in before.
	$printing = $found;

    } else {
        print if $printing;
    }

    # Cleanup at the end of each file
    if (eof) {
	# Finish nesting subsections
	# I.e., if we had any subsections in this section, end subsectioning
	foreach my $level (1 .. $#sec_counter) {
	    if ($sec_counter[$level]) {
		print "\\end_deeper\n";
	    }
	}
	# So that on next file we check whether it's an article or book
        @matches = ();
    }

}

print "\n\\the_end\n";

# Make a (nested) section number
# Input is current section numbers
sub section_number {
    my $number = shift; # Highest level section counter
    foreach my $index (@_) {
        if ($index) {
	    $number .= ".$index";
	} else {
	    # Done creating the number
	    last;
	}
    }
    return $number;
}
